import time

import shortuuid

from django_unicorn.utils import generate_checksum


def post_and_get_response(
    client,
    *,
    url="",
    data=None,
    action_queue=None,
    component_id=None,
    hash=None,  # noqa: A002
    return_response=False,
):
    if not data:
        data = {}
    if not action_queue:
        action_queue = []
    if not component_id:
        component_id = shortuuid.uuid()[:8]

    message = {
        "actionQueue": action_queue,
        "data": data,
        "checksum": generate_checksum(data),
        "id": component_id,
        "epoch": time.time(),
        "hash": hash,
    }

    response = client.post(
        url,
        message,
        content_type="application/json",
    )

    if return_response:
        return response

    try:
        return response.json()
    except TypeError:
        # Return the regular response if no JSON for HttpResponseNotModified
        return response



import datetime
from dataclasses import dataclass
from typing import get_type_hints as typing_get_type_hints

from pydantic import BaseModel

from django_unicorn.components import UnicornView
from django_unicorn.typer import cast_attribute_value, cast_value, get_type_hints
from example.coffee.models import Flavor


def test_get_type_hints():
    def test_func(input_str: str):
        return input_str

    expected = {"input_str": str}
    actual = get_type_hints(test_func)
    assert actual == expected


def test_get_type_hints_missing_type_hints():
    def test_func(input_str):
        return input_str

    expected = {}
    actual = get_type_hints(test_func)
    assert actual == expected


def test_get_type_hints_gh_639():
    class MyComponentView(UnicornView):
        a_date: datetime.date

    actual = get_type_hints(MyComponentView(component_name="test", component_id="test_get_type_hints_gh_639"))

    assert "a_date" in actual
    assert actual["a_date"] == datetime.date


class TestClass:
    integer: int
    boolean: bool


def test_cast_attribute_value_int():
    expected = 1
    actual = cast_attribute_value(TestClass(), "integer", "1")

    assert expected == actual


def test_cast_attribute_value_bool_true():
    expected = True
    actual = cast_attribute_value(TestClass(), "boolean", "True")

    assert expected == actual


def test_cast_attribute_value_bool_false():
    expected = False
    actual = cast_attribute_value(TestClass(), "boolean", "False")

    assert expected == actual


def test_cast_attribute_value_bool_invalid():
    expected = False
    actual = cast_attribute_value(TestClass(), "boolean", "asdf")

    assert expected == actual


class ExampleClass:
    a_model: Flavor
    optional_model: Flavor | None


def test_cast_value_model_none():
    example_class = ExampleClass()
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["a_model"]

    actual = cast_value(type_hint, None)

    assert actual is None


def test_cast_value_model_dict():
    example_class = ExampleClass()
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["a_model"]

    actual = cast_value(type_hint, {"id": 1})

    assert actual == {"id": 1}


def test_cast_value_optional():
    type_hints = typing_get_type_hints(ExampleClass())
    type_hint = type_hints["optional_model"]

    actual = cast_value(type_hint, None)

    assert actual is None


def test_cast_value_model_int():
    example_class = ExampleClass()
    example_class.a_model = Flavor()

    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["optional_model"]

    actual = cast_value(type_hint, 1)

    assert actual == 1


@dataclass
class DataClass:
    name: str


class PydanticBaseModel(BaseModel):
    name: str


class AnotherExampleClass:
    data: DataClass
    list_data: list[DataClass]
    pydantic_data: PydanticBaseModel
    pydantic_list_data: list[PydanticBaseModel]


def test_cast_value_dataclass():
    example_class = AnotherExampleClass()
    test_data = DataClass(name="foo")
    example_class.data = test_data
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["data"]
    actual = cast_value(type_hint, {"name": "foo"})
    assert actual == test_data


def test_cast_value_pydantic():
    example_class = AnotherExampleClass()
    test_data = PydanticBaseModel(name="foo")
    example_class.pydantic_data = test_data
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["pydantic_data"]
    actual = cast_value(type_hint, {"name": "foo"})
    assert actual == test_data


def test_cast_value_list_dataclass():
    example_class = AnotherExampleClass()
    test_data = DataClass(name="foo")
    example_class.list_data = [test_data]
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["list_data"]
    actual = cast_value(type_hint, [{"name": "foo"}])
    assert actual == [test_data]


def test_cast_value_list_pydantic():
    example_class = AnotherExampleClass()
    test_data = PydanticBaseModel(name="foo")
    example_class.pydantic_list_data = [test_data]
    type_hints = typing_get_type_hints(example_class)
    type_hint = type_hints["pydantic_list_data"]
    actual = cast_value(type_hint, [{"name": "foo"}])
    assert actual == [test_data]



from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.test import RequestFactory

from django_unicorn.components.unicorn_view import UnicornView, construct_component


class RedirectView(UnicornView):
    def mount(self):
        return redirect("/somewhere-else")


class ResponseView(UnicornView):
    def mount(self):
        return HttpResponse("Custom Content")


def test_mount_redirect_direct_view():
    request = RequestFactory().get("/")
    view = RedirectView.as_view()
    response = view(request)

    assert isinstance(response, HttpResponseRedirect)
    assert response.url == "/somewhere-else"


def test_mount_response_direct_view():
    request = RequestFactory().get("/")
    view = ResponseView.as_view()
    response = view(request)

    assert response.status_code == 200
    assert response.content.decode() == "Custom Content"


def test_mount_redirect_template_tag():
    # For now, let's verify construct_component captures it (once implemented)
    component = construct_component(RedirectView, "123", "test", "", None, None, [])

    # This attribute doesn't exist yet, but it's part of the plan
    assert hasattr(component, "_mount_result")
    assert isinstance(component._mount_result, HttpResponseRedirect)
    assert component._mount_result.url == "/somewhere-else"


def test_render_redirect_template_tag():
    component = construct_component(RedirectView, "123", "test", "", None, None, [])
    html = component.render()

    # Should be a script tag with generic redirect
    assert "<script>" in html
    assert "window.location.href = '/somewhere-else'" in html



from django.utils.timezone import now

from django_unicorn.components import UnicornView


class TextInputsView(UnicornView):
    name = "World"
    testing_xss = "Whatever </script> <script>alert('uh oh')</script>"

    def set_name(self, name=None):
        if name:
            self.name = name
        else:
            self.name = "Universe"

        return f"{self.name} - {now().second}"



from django_unicorn.typer import get_type_hints


class Parent:
    name: str


class Child(Parent):
    age: int


def test_get_type_hints_inheritance():
    """
    Verify that get_type_hints returns type hints from parent classes
    when called with an instance of a subclass.
    """
    instance = Child()
    type_hints = get_type_hints(instance)

    assert "name" in type_hints
    assert type_hints["name"] is str
    assert "age" in type_hints
    assert type_hints["age"] is int


if __name__ == "__main__":
    test_get_type_hints_inheritance()



import pytest

from django_unicorn.call_method_parser import InvalidKwargError, parse_kwarg


def test_kwargs_string():
    expected = {"test": "1"}
    actual = parse_kwarg("test='1'")

    assert actual == expected
    assert isinstance(actual["test"], str)
    assert actual["test"] == "1"


def test_kwargs_int():
    expected = {"test": 2}
    actual = parse_kwarg("test=2")

    assert actual == expected
    assert isinstance(actual["test"], int)
    assert actual["test"] == 2


def test_kwargs_invalid_startswith_doublequote():
    with pytest.raises(InvalidKwargError) as e:
        parse_kwarg('"test"=2')

    assert e.type == InvalidKwargError


def test_kwargs_invalid_startswith_singlequote():
    with pytest.raises(InvalidKwargError) as e:
        parse_kwarg("'test'=2")

    assert e.type == InvalidKwargError


def test_kwargs_invalid_no_equal_sign():
    with pytest.raises(InvalidKwargError) as e:
        parse_kwarg("test")

    assert e.type == InvalidKwargError


def test_kwargs_invalid_internal_doublequote():
    with pytest.raises(InvalidKwargError) as e:
        parse_kwarg('te"st=1')

    assert e.type == InvalidKwargError


def test_kwargs_invalid_internal_singlequote():
    with pytest.raises(InvalidKwargError) as e:
        parse_kwarg("te'st=1")

    assert e.type == InvalidKwargError


def test_kwargs_skip_unparseable_value():
    expected = {"test": "some_context_variable"}
    actual = parse_kwarg("test=some_context_variable")

    assert actual == expected
    assert isinstance(actual["test"], str)
    assert actual["test"] == "some_context_variable"


def test_kwargs_raise_unparseable_value():
    with pytest.raises(ValueError) as e:
        parse_kwarg("test=some_context_variable", raise_if_unparseable=True)

    assert e.type is ValueError



from django_unicorn.components import UnicornView
from example.coffee.models import Flavor


class TableView(UnicornView):
    name = "Coffee Flavors"
    original_name = None
    flavors = Flavor.objects.none()
    is_editing = False
    favorite_count = 0
    show_filter = False

    def edit(self):
        self.is_editing = True
        self.original_name = self.name

    def save(self):
        self.is_editing = False

    def cancel(self):
        if self.original_name:
            self.name = self.original_name
            self.original_name = None

        self.is_editing = False

    def mount(self):
        self.load_table()

    def load_table(self):
        self.flavors = Flavor.objects.select_related("favorite").all()[10:20]
        self.favorite_count = sum([1 for f in self.flavors if hasattr(f, "favorite") and f.favorite.is_favorite])

        def set_unedit(c):
            if hasattr(c, "is_editing"):
                c.is_editing = False
            for cc in c.children:
                set_unedit(cc)

        for child in self.children:
            set_unedit(child)



from datetime import date, datetime, time, timedelta
from types import MappingProxyType
from uuid import UUID, uuid4

import pytest

from django_unicorn.components import UnicornView
from django_unicorn.views.action_parsers.call_method import _call_method_name
from example.coffee.models import Flavor


class CustomClass:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class InvalidCustomClass:
    pass


class FakeComponent(UnicornView):
    def save(self):
        return 1

    def save_with_arg(self, id):  # noqa: A002
        return 1 + id

    def save_with_kwarg(self, id=0):  # noqa: A002
        return 1 + id

    def save_with_model(self, model: Flavor):
        assert issubclass(type(model), Flavor), "model arg should be a `Flavor` model"
        return model.pk

    def save_with_union(self, id: int | str):  # noqa: A002
        assert isinstance(id, int | str)
        return id

    def get_datetime_without_type_hint(self, _datetime):
        return _datetime

    def get_datetime(self, _datetime: datetime):
        return _datetime

    def get_date(self, _date: date):
        return _date

    def get_datetime_and_date(self, _datetime: datetime, _date: date):
        return (_datetime, _date)

    def get_duration(self, _duration: timedelta):
        return _duration

    def get_time(self, _time: time):
        return _time

    def get_duration_and_time(self, _duration: timedelta | None = None, _time: time | None = None):
        return (_duration, _time)

    def get_uuid(self, _uuid: UUID):
        return _uuid

    def get_custom_class(self, _cls: CustomClass):
        return _cls

    def get_invalid_custom_class(self, _cls: InvalidCustomClass):
        return _cls


def test_call_method_name_missing():
    component = FakeComponent(component_name="test", component_id="test_call_method_name_missing")
    actual = _call_method_name(component, "save_missing", args=(), kwargs={})

    assert actual is None


def test_call_method_name_no_params():
    expected = 1

    component = FakeComponent(component_name="test", component_id="test_call_method_name_no_params")
    actual = _call_method_name(component, "save", args=(), kwargs={})

    assert actual == expected


def test_call_method_name_with_arg():
    expected = 2

    component = FakeComponent(component_name="test", component_id="test_call_method_name_with_arg")
    actual = _call_method_name(component, "save_with_arg", args=(1,), kwargs={})

    assert actual == expected


def test_call_method_name_with_kwarg():
    expected = 3

    component = FakeComponent(component_name="test", component_id="test_call_method_name_with_kwarg")
    actual = _call_method_name(component, "save_with_kwarg", args=(), kwargs=MappingProxyType({"id": 2}))

    assert actual == expected


@pytest.mark.django_db
def test_call_method_name_arg_with_model_type_annotation():
    flavor = Flavor()
    flavor.save()

    component = FakeComponent(
        component_name="test", component_id="test_call_method_name_arg_with_model_type_annotation"
    )
    actual = _call_method_name(component, "save_with_model", args=(flavor.pk,), kwargs={})

    assert actual == flavor.pk


@pytest.mark.django_db
def test_call_method_name_arg_with_model_type_annotation_multiple():
    flavor_one = Flavor()
    flavor_one.save()

    flavor_two = Flavor()
    flavor_two.save()

    assert flavor_one.pk != flavor_two.pk

    component = FakeComponent(
        component_name="test", component_id="test_call_method_name_arg_with_model_type_annotation_multiple"
    )
    actual = _call_method_name(component, "save_with_model", args=(flavor_one.pk,), kwargs={})
    assert actual == flavor_one.pk

    # second call
    actual = _call_method_name(component, "save_with_model", args=(flavor_two.pk,), kwargs={})
    assert actual == flavor_two.pk

    # third call
    actual = _call_method_name(component, "save_with_model", args=(flavor_one.pk,), kwargs={})
    assert actual == flavor_one.pk

    # fourth call
    actual = _call_method_name(component, "save_with_model", args=(flavor_one.pk,), kwargs={})
    assert actual == flavor_one.pk


def _get_actual(method_name: str, args=None, kwargs=None):
    component = FakeComponent(component_name="test", component_id="test-call-method-name-tests")

    if args is None:
        args = ()

    if kwargs is None:
        kwargs = {}

    return _call_method_name(
        component,
        method_name,
        args=args,
        kwargs=kwargs,
    )


@pytest.mark.django_db
def test_call_method_name_kwarg_with_model_type_annotation():
    flavor = Flavor()
    flavor.save()

    actual = _get_actual("save_with_model", kwargs=MappingProxyType({"pk": flavor.pk}))

    assert actual == flavor.pk


def test_call_method_name_with_kwarg_with_union_and_int():
    actual = _get_actual("save_with_union", kwargs=MappingProxyType({"id": 2}))

    assert isinstance(actual, int)


def test_call_method_name_with_kwarg_with_union_and_str():
    actual = _get_actual("save_with_union", kwargs=MappingProxyType({"id": "2"}))

    assert isinstance(actual, int)


def test_call_method_name_without_type_hint():
    actual = _get_actual("get_datetime_without_type_hint", args=["2020-09-12T01:01:01"])

    assert isinstance(actual, str)


def test_call_method_name_with_datetime_type_hint():
    actual = _get_actual("get_datetime", args=["2020-09-12T01:01:01"])

    assert isinstance(actual, datetime)


def test_call_method_name_with_date_type_hint():
    actual = _get_actual("get_date", args=["2020-09-12"])

    assert isinstance(actual, date)


def test_call_method_name_with_time_type_hint():
    actual = _get_actual("get_time", args=["01:01:01"])

    assert isinstance(actual, time)


def test_call_method_name_with_datetime_and_date_type_hint():
    actual = _get_actual("get_datetime_and_date", args=["2020-09-12T01:01:01", "2020-09-12"])

    assert isinstance(actual[0], datetime)
    assert isinstance(actual[1], date)


def test_call_method_name_with_duration_type_hint():
    actual = _get_actual("get_duration", args=["3"])

    assert isinstance(actual, timedelta)


def test_call_method_name_with_duration_and_time_type_hint():
    actual = _get_actual("get_duration_and_time", kwargs={"_duration": "3", "_time": "1:01:01"})

    assert isinstance(actual[0], timedelta)
    assert isinstance(actual[1], time)


def test_call_method_name_with_datetime_as_epoch_type_hint():
    actual = _get_actual("get_datetime", kwargs={"_datetime": 1691499534})

    assert isinstance(actual, datetime)


def test_call_method_name_with_date_as_epoch_type_hint():
    actual = _get_actual("get_date", kwargs={"_date": 1691499534})

    assert isinstance(actual, date)
    assert actual == date(2023, 8, 8)


def test_call_method_name_with_uuid_type_hint():
    actual = _get_actual("get_uuid", kwargs={"_uuid": str(uuid4())})

    assert isinstance(actual, UUID)


def test_call_method_name_with_custom_class_type_hint():
    actual = _get_actual("get_custom_class", kwargs={"_cls": "test"})

    assert isinstance(actual, CustomClass)
    assert actual.args[0] == "test"


def test_call_method_name_with_invalid_custom_class_type_hint():
    with pytest.raises(TypeError):
        _get_actual("get_invalid_custom_class", kwargs={"_cls": "test"})



# Generated by Django 3.1.7 on 2021-04-11 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('coffee', '0003_auto_20210128_0140'),
    ]

    operations = [
        migrations.CreateModel(
            name='Taste',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('flavor', models.ManyToManyField(to='coffee.Flavor')),
            ],
        ),
        migrations.CreateModel(
            name='Origin',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('flavor', models.ManyToManyField(related_name='origins', to='coffee.Flavor')),
            ],
        ),
    ]



from datetime import datetime

from django_unicorn.components import UnicornView

from ..forms import ValidationForm


class ValidationView(UnicornView):
    form_class = ValidationForm

    text = "hello"
    number = ""
    date_time = datetime(2020, 9, 13, 17, 45, 14)
    email = ""

    def set_text_no_validation(self):
        self.text = "no validation"

    def set_text_with_validation(self):
        self.text = "validation"
        self.validate()

    def set_number(self, number):
        self.number = number



from django_unicorn.components.unicorn_view import convert_to_dash_case


def test_convert_to_dash_case():
    expected = "hello-world"
    actual = convert_to_dash_case("hello_world")

    assert expected == actual



from django_unicorn.components.unicorn_view import convert_to_pascal_case


def test_convert_to_pascal_case():
    expected = "HelloWorld"
    actual = convert_to_pascal_case("hello-world")

    assert expected == actual



import json
import uuid
from datetime import timedelta
from decimal import Decimal
from types import MappingProxyType
from typing import cast

import pytest
from django.db import models
from django.db.models import Model
from django.utils.timezone import now
from pydantic import BaseModel

from django_unicorn import serializer
from django_unicorn.serializer import InvalidFieldAttributeError, InvalidFieldNameError
from django_unicorn.utils import dicts_equal
from example.coffee.models import Flavor, NewFlavor, Taste


class SimpleTestModel(models.Model):
    name = models.CharField(max_length=10)

    class Meta:
        app_label = "tests"


class SubclassSimpleTestModel(SimpleTestModel):
    subclass_name = models.CharField(max_length=10)

    class Meta:
        app_label = "tests"


class ComplicatedTestModel(models.Model):
    name = models.CharField(max_length=10)
    datetime = models.DateTimeField(auto_now=True, auto_now_add=True)
    float_value = models.FloatField(default=0.583)
    decimal_value = models.DecimalField(max_digits=10, decimal_places=2, default=0.984)
    uuid = models.UUIDField(default=uuid.uuid4)
    date = models.DateField(auto_now=True, auto_now_add=True)
    time = models.TimeField(auto_now=True, auto_now_add=True)
    duration = models.DurationField(default="-1 day, 19:00:00")

    class Meta:
        app_label = "tests"


class SubclassComplicatedTestModel(ComplicatedTestModel):
    subclass_name = models.CharField(max_length=10)

    class Meta:
        app_label = "tests"


class ForeignKeyTestModel(models.Model):
    name = models.CharField(max_length=10)
    parent = models.ForeignKey("self", blank=True, null=True, on_delete=models.SET_NULL)

    class Meta:
        app_label = "tests"


class SubclassForeignKeyTestModel(ForeignKeyTestModel):
    subclass_name = models.CharField(max_length=10)

    class Meta:
        app_label = "tests"


class ManyToManyTestModel(models.Model):
    name = models.CharField(max_length=10)
    parents = models.ManyToManyField("self")

    class Meta:
        app_label = "tests"


class SubclassManyToManyTestModel(ManyToManyTestModel):
    subclass_name = models.CharField(max_length=10)

    class Meta:
        app_label = "tests"


def assert_dicts(expected: dict, serialized_dumps: str) -> None:
    actual = json.loads(serialized_dumps)

    assert dicts_equal(expected, actual)


def test_int():
    expected = '{"name":123}'
    actual = serializer.dumps({"name": 123})

    assert expected == actual


def test_decimal():
    expected = '{"name":"123.1"}'
    actual = serializer.dumps({"name": Decimal("123.1")})

    assert expected == actual


def test_mapping_proxy_type():
    expected = '{"name":{"id":"123.1"}}'
    actual = serializer.dumps({"name": MappingProxyType({"id": Decimal("123.1")})})

    assert expected == actual


def test_string():
    expected = '{"name":"abc"}'
    actual = serializer.dumps({"name": "abc"})

    assert expected == actual


def test_list():
    expected = '{"name":["abc","def"]}'
    actual = serializer.dumps(
        {
            "name": [
                "abc",
                "def",
            ]
        }
    )

    assert expected == actual


def test_simple_model():
    expected = {"name": "abc", "pk": 1}

    simple_test_model = SimpleTestModel(id=1, name="abc")
    actual = serializer.dumps(simple_test_model)

    assert_dicts(expected, actual)


def test_subclass_simple_model():
    expected = {"subclass_name": "def", "pk": 2, "name": "abc"}

    subclass_simple_test_model = SubclassSimpleTestModel(id=2, name="abc", subclass_name="def")
    actual = serializer.dumps(subclass_simple_test_model)

    assert_dicts(expected, actual)


def test_complicated_model():
    now_dt = now()
    duration = timedelta(days=-1, seconds=68400)

    model = ComplicatedTestModel(
        id=2,
        name="abc",
        datetime=now_dt,
        date=now_dt.date(),
        time=now_dt.time(),
        duration=duration,
    )

    expected = {
        "date": str(model.date),
        "datetime": model.datetime.isoformat(timespec="milliseconds").replace("+00:00", "Z"),
        "decimal_value": "0.984",
        "duration": "-1 19:00:00",
        "float_value": "0.583",
        "name": "abc",
        "pk": 2,
        "time": str(model.time)[:-3],
        "uuid": str(model.uuid),
    }

    actual = serializer.dumps(model)

    assert_dicts(expected, actual)


def test_subclass_complicated_model():
    now_dt = now()
    duration = timedelta(days=-1, seconds=68400)

    model = SubclassComplicatedTestModel(
        id=2,
        name="abc",
        subclass_name="def",
        datetime=now_dt,
        date=now_dt.date(),
        time=now_dt.time(),
        duration=duration,
    )

    expected = {
        "subclass_name": "def",
        "pk": 2,
        "name": "abc",
        "datetime": model.datetime.isoformat(timespec="milliseconds").replace("+00:00", "Z"),
        "float_value": "0.583",
        "decimal_value": "0.984",
        "uuid": str(model.uuid),
        "date": str(model.date),
        "time": str(model.time)[:-3],
        "duration": "-1 19:00:00",
    }

    actual = serializer.dumps(model)

    assert_dicts(expected, actual)


def test_model_with_timedelta(db):  # noqa: ARG001
    now_dt = now()
    duration = timedelta(days=-1, seconds=68400)

    flavor = Flavor(
        id=8,
        name="name1",
        datetime=now_dt,
        date=now_dt.date(),
        time=now_dt.time(),
        duration=duration,
    )

    expected = {
        "flavor": {
            "name": "name1",
            "label": "",
            "parent": None,
            "float_value": None,
            "decimal_value": None,
            "uuid": str(flavor.uuid),
            "date": str(now_dt.date()),
            "datetime": now_dt.isoformat(timespec="milliseconds").replace("+00:00", "Z"),
            "time": str(now_dt.time())[:-3],
            "duration": "-1 19:00:00",
            "pk": 8,
            "taste_set": [],
            "origins": [],
        }
    }

    actual = serializer.dumps({"flavor": flavor})

    assert_dicts(expected, actual)


def test_model_with_datetime_as_string(db):  # noqa: ARG001
    datetime = now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
    flavor = Flavor(name="name1", datetime=datetime)

    expected = {
        "flavor": {
            "name": "name1",
            "label": "",
            "parent": None,
            "float_value": None,
            "decimal_value": None,
            "uuid": str(flavor.uuid),
            "date": None,
            "datetime": datetime.replace(".000", ""),
            "time": None,
            "duration": None,
            "pk": None,
            "taste_set": [],
            "origins": [],
        }
    }

    actual = serializer.dumps({"flavor": flavor})

    assert_dicts(expected, actual)


def test_model_with_time_as_string(db):  # noqa: ARG001
    time = now().strftime("%H:%M:%S.%f")[:-3]
    flavor = Flavor(name="name1", time=time)

    expected = {
        "flavor": {
            "name": "name1",
            "label": "",
            "parent": None,
            "float_value": None,
            "decimal_value": None,
            "uuid": str(flavor.uuid),
            "date": None,
            "datetime": None,
            "time": time,
            "duration": None,
            "pk": None,
            "taste_set": [],
            "origins": [],
        }
    }

    actual = serializer.dumps({"flavor": flavor})

    assert_dicts(expected, actual)


def test_model_with_duration_as_string(db):  # noqa: ARG001
    duration = "-1 day, 19:00:00"
    flavor = Flavor(name="name1", duration=duration)

    expected = {
        "flavor": {
            "name": "name1",
            "label": "",
            "parent": None,
            "float_value": None,
            "decimal_value": None,
            "uuid": str(flavor.uuid),
            "date": None,
            "datetime": None,
            "time": None,
            "duration": "-1 19:00:00",
            "pk": None,
            "taste_set": [],
            "origins": [],
        }
    }

    actual = serializer.dumps({"flavor": flavor})

    assert_dicts(expected, actual)


def test_model_foreign_key():
    test_model_one = ForeignKeyTestModel(id=1, name="abc")
    test_model_two = ForeignKeyTestModel(id=2, name="def", parent=test_model_one)
    expected = {"test_model_two": {"name": "def", "parent": 1, "pk": 2}}

    actual = serializer.dumps({"test_model_two": test_model_two})

    assert_dicts(expected, actual)


def test_model_foreign_key_subclass():
    test_model_one = ForeignKeyTestModel(id=1, name="abc")
    test_model_two = SubclassForeignKeyTestModel(id=2, name="def", subclass_name="ghi", parent=test_model_one)
    expected = {"name": "def", "parent": 1, "pk": 2, "subclass_name": "ghi"}

    actual = serializer.dumps(test_model_two)

    assert_dicts(expected, actual)


@pytest.mark.django_db
def test_model_inherited_model_with_many_to_many_and_foreign_key():
    flavor_one = Flavor(id=3, name="name0")
    flavor_one.save()

    flavor_two = NewFlavor(id=4, new_name="new_name_1", name="name_1", parent=flavor_one)
    flavor_two.save()

    taste1 = Taste(name="Bitter1")
    taste1.save()
    flavor_two.taste_set.add(taste1)  # type: ignore

    expected = {
        "new_name": "new_name_1",
        "pk": 4,
        "taste_set": [1],
        "origins": [],
        "name": "name_1",
        "label": "",
        "parent": 3,
        "float_value": "null",
        "decimal_value": "null",
        "uuid": str(flavor_two.uuid),
        "datetime": "null",
        "date": "null",
        "time": "null",
        "duration": "null",
    }

    actual = serializer.dumps(flavor_two)

    assert_dicts(expected, actual)


def test_model_foreign_key_recursive_parent():
    test_model_one = ForeignKeyTestModel(id=1, name="abc")
    test_model_two = ForeignKeyTestModel(id=2, name="def", parent=test_model_one)
    test_model_one.parent = test_model_two
    expected = '{"test_model_two":{"name":"def","parent":1,"pk":2}}'

    actual = serializer.dumps({"test_model_two": test_model_two})

    assert expected == actual


@pytest.mark.django_db
def test_model_many_to_many(django_assert_num_queries):
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()

    taste1 = Taste(name="Bitter1")
    taste1.save()
    taste2 = Taste(name="Bitter2")
    taste2.save()
    taste3 = Taste(name="Bitter3")
    taste3.save()

    flavor_one.taste_set.add(taste1)  # type: ignore
    flavor_one.taste_set.add(taste2)  # type: ignore
    flavor_one.taste_set.add(taste3)  # type: ignore

    with django_assert_num_queries(2):
        actual = json.loads(serializer.dumps(flavor_one))

    expected = {
        "name": "name1",
        "label": "label1",
        "parent": None,
        "float_value": None,
        "decimal_value": None,
        "uuid": str(flavor_one.uuid),
        "datetime": None,
        "date": None,
        "time": None,
        "duration": None,
        "pk": 1,
        "taste_set": [taste1.pk, taste2.pk, taste3.pk],
        "origins": [],
    }

    assert dicts_equal(expected, actual)


@pytest.mark.django_db
def test_model_many_to_many_with_excludes(django_assert_num_queries):
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()

    taste1 = Taste(name="Bitter1")
    taste1.save()
    taste2 = Taste(name="Bitter2")
    taste2.save()
    taste3 = Taste(name="Bitter3")
    taste3.save()

    flavor_one.taste_set.add(taste1)  # type: ignore
    flavor_one.taste_set.add(taste2)  # type: ignore
    flavor_one.taste_set.add(taste3)  # type: ignore

    flavor_one = Flavor.objects.prefetch_related("taste_set", "origins").get(pk=flavor_one.pk)

    # This shouldn't make any database calls because of the prefetch_related
    with django_assert_num_queries(0):
        actual = serializer.dumps(
            {"flavor": flavor_one},
            exclude_field_attributes=(
                "flavor.taste_set",
                "flavor.origins",
            ),
        )
        actual = json.loads(actual)

    expected = {
        "flavor": {
            "name": "name1",
            "label": "label1",
            "parent": None,
            "float_value": None,
            "decimal_value": None,
            "uuid": str(flavor_one.uuid),
            "datetime": None,
            "date": None,
            "time": None,
            "duration": None,
            "pk": 1,
        }
    }

    assert dicts_equal(expected, actual)


@pytest.mark.django_db
def test_dumps_queryset(db):  # noqa: ARG001
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()

    flavor_two = Flavor(name="name2", label="label2", parent=flavor_one)
    flavor_two.save()

    flavors = Flavor.objects.all()

    expected = {
        "flavors": [
            {
                "name": "name1",
                "label": "label1",
                "parent": None,
                "float_value": None,
                "decimal_value": None,
                "uuid": str(flavor_one.uuid),
                "date": None,
                "datetime": None,
                "time": None,
                "duration": None,
                "pk": 1,
                "taste_set": [],
                "origins": [],
            },
            {
                "name": "name2",
                "label": "label2",
                "parent": 1,
                "float_value": None,
                "decimal_value": None,
                "uuid": str(flavor_two.uuid),
                "date": None,
                "datetime": None,
                "time": None,
                "duration": None,
                "pk": 2,
                "taste_set": [],
                "origins": [],
            },
        ]
    }

    actual = json.loads(serializer.dumps({"flavors": flavors}))
    assert dicts_equal(expected, actual)


def test_get_model_dict():
    flavor_one = Flavor(name="name1", label="label1")
    actual = serializer._get_model_dict(flavor_one)

    expected = {
        "pk": None,
        "name": "name1",
        "label": "label1",
        "parent": None,
        "decimal_value": None,
        "float_value": None,
        "uuid": str(flavor_one.uuid),
        "date": None,
        "datetime": None,
        "time": None,
        "duration": None,
        "taste_set": [],
        "origins": [],
    }

    assert dicts_equal(expected, actual)


@pytest.mark.django_db
def test_get_model_dict_many_to_many_is_referenced(django_assert_num_queries):
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()

    taste1 = Taste(name="Bitter")
    taste1.save()
    taste2 = Taste(name="Bitter2")
    taste2.save()
    taste3 = Taste(name="Bitter3")
    taste3.save()

    flavor_one.taste_set.add(taste1)  # type: ignore
    flavor_one.taste_set.add(taste2)  # type: ignore
    flavor_one.taste_set.add(taste3)  # type: ignore

    expected = {
        "pk": 1,
        "name": "name1",
        "label": "label1",
        "parent": None,
        "decimal_value": None,
        "float_value": None,
        "uuid": str(flavor_one.uuid),
        "date": None,
        "datetime": None,
        "time": None,
        "duration": None,
        "taste_set": [taste1.pk, taste2.pk, taste3.pk],
        "origins": [],
    }

    flavor_one = Flavor.objects.filter(id=flavor_one.id).first()  # type: ignore

    with django_assert_num_queries(2):
        actual = serializer._get_model_dict(cast(Model, flavor_one))

    assert dicts_equal(expected, actual)


@pytest.mark.django_db
def test_get_model_dict_many_to_many_is_referenced_prefetched(
    django_assert_num_queries,
):
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()

    taste1 = Taste(name="Bitter")
    taste1.save()
    taste2 = Taste(name="Bitter2")
    taste2.save()
    taste3 = Taste(name="Bitter3")
    taste3.save()

    flavor_one.taste_set.add(taste1)  # type: ignore
    flavor_one.taste_set.add(taste2)  # type: ignore
    flavor_one.taste_set.add(taste3)  # type: ignore

    expected = {
        "pk": 1,
        "name": "name1",
        "label": "label1",
        "parent": None,
        "decimal_value": None,
        "float_value": None,
        "uuid": str(flavor_one.uuid),
        "date": None,
        "datetime": None,
        "time": None,
        "duration": None,
        "taste_set": [taste1.pk, taste2.pk, taste3.pk],
        "origins": [],
    }

    flavor_one = Flavor.objects.prefetch_related("taste_set").filter(pk=flavor_one.pk).first()

    # prefetch_related should reduce the database calls
    with django_assert_num_queries(1):
        actual = serializer._get_model_dict(cast(Model, flavor_one))

    assert dicts_equal(expected, actual)


@pytest.mark.django_db
def test_get_model_dict_many_to_many_references_model():
    taste = Taste(name="Bitter")
    taste.save()
    flavor_one = Flavor(name="name1", label="label1")
    flavor_one.save()
    flavor_one.taste_set.add(taste)  # type: ignore
    actual = serializer._get_model_dict(taste)

    expected = {"name": taste.name, "flavor": [flavor_one.pk], "pk": taste.pk}

    assert dicts_equal(expected, actual)


def test_float():
    expected = '{"name":"0.0"}'
    actual = serializer.dumps({"name": 0.0})

    assert expected == actual


def test_dict_float():
    expected = '{"name":{"another":"0.0"}}'
    actual = serializer.dumps({"name": {"another": 0.0}})

    assert expected == actual


def test_list_float():
    expected = '{"name":[1,2,"0.0"]}'
    actual = serializer.dumps({"name": [1, 2, 0.0]})

    assert expected == actual


def test_nested_list_float():
    expected = '{"name":{"blob":[1,2,"0.0"]}}'
    actual = serializer.dumps({"name": {"blob": [1, 2, 0.0]}})

    assert expected == actual


def test_nested_list_float_complicated():
    expected = '{"another":[{"great":"1.0","ok":["1.6","0.0",4]}],"more":["1.9",2,5],"name":{"blob":[1,2,"0.0"]}}'
    actual = serializer.dumps(
        {
            "name": {"blob": [1, 2, 0.0]},
            "more": [1.9, 2, 5],
            "another": [{"great": 1.0, "ok": [1.6, 0.0, 4]}],
        }
    )

    assert expected == actual


def test_nested_list_float_less_complicated():
    expected = '{"another":[{"great":"1.0","ok":["1.6","0.0",4]}]}'
    actual = serializer.dumps(
        {
            "another": [{"great": 1.0, "ok": [1.6, 0.0, 4]}],
        }
    )

    assert expected == actual


def test_pydantic():
    class Book(BaseModel):
        title: str = "The Grapes of Wrath"
        author: str = "John Steinbeck"

    expected = '{"author":"John Steinbeck","title":"The Grapes of Wrath"}'
    actual = serializer.dumps(Book())

    assert expected == actual


def test_exclude_field_attributes():
    expected = '{"book":{"title":"The Grapes of Wrath"}}'

    actual = serializer.dumps(
        {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
        exclude_field_attributes=("book.author",),
    )

    assert expected == actual


def test_exclude_field_attributes_no_fix_floats():
    expected = '{"book":{"title":"The Grapes of Wrath"}}'

    actual = serializer.dumps(
        {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
        fix_floats=False,
        exclude_field_attributes=("book.author",),
    )

    assert expected == actual


def test_exclude_field_attributes_nested():
    expected = '{"classic":{"book":{"title":"The Grapes of Wrath"}}}'

    actual = serializer.dumps(
        {"classic": {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}}},
        exclude_field_attributes=("classic.book.author",),
    )

    assert expected == actual


def test_exclude_field_attributes_invalid_field_attribute():
    expected = '{"book":{"author":"John Steinbeck","title":"The Grapes of Wrath"}}'

    actual = serializer.dumps(
        {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
        exclude_field_attributes=("blob",),
    )

    assert expected == actual


def test_exclude_field_attributes_empty_string():
    expected = '{"book":{"author":"John Steinbeck","title":"The Grapes of Wrath"}}'

    actual = serializer.dumps(
        {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
        exclude_field_attributes=("",),
    )

    assert expected == actual


def test_exclude_field_attributes_none():
    expected = '{"book":{"author":"John Steinbeck","title":"The Grapes of Wrath"}}'

    actual = serializer.dumps(
        {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
        exclude_field_attributes=None,
    )

    assert expected == actual


def test_exclude_field_attributes_invalid_name():
    with pytest.raises(InvalidFieldNameError) as e:
        serializer.dumps(
            {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
            exclude_field_attributes=("blob.monster",),
        )

    assert e.exconly() == "django_unicorn.serializer.InvalidFieldNameError: Cannot resolve 'blob'. Choices are: book"


def test_exclude_field_attributes_invalid_attribute():
    with pytest.raises(InvalidFieldAttributeError) as e:
        serializer.dumps(
            {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
            exclude_field_attributes=("book.blob",),
        )

    assert (
        e.exconly()
        == "django_unicorn.serializer.InvalidFieldAttributeError: Cannot resolve 'blob'. Choices on 'book' are: title, author"  # noqa: E501
    )


def test_exclude_field_attributes_invalid_type():
    with pytest.raises(AssertionError):
        serializer.dumps(
            {"book": {"title": "The Grapes of Wrath", "author": "John Steinbeck"}},
            exclude_field_attributes=("book.blob"),  # type: ignore
        )


def test_dictionary_with_int_keys_as_strings():
    # Browsers will sort a dictionary that has stringified integers as if they the keys
    # were integers which messes up checksum calculation
    expected = '{"1":"a","2":"b","4":"c","11":"d","24":"e"}'

    actual = serializer.dumps(
        {
            "11": "d",
            "4": "c",
            "1": "a",
            "24": "e",
            "2": "b",
        }
    )

    assert expected == actual


def test_dictionary_with_int_keys_as_strings_no_sort():
    expected = '{"11":"d","4":"c","1":"a","24":"e","2":"b"}'

    actual = serializer.dumps(
        {
            "11": "d",
            "4": "c",
            "1": "a",
            "24": "e",
            "2": "b",
        },
        sort_dict=False,
    )

    assert expected == actual



import logging
import re
from collections import deque

import orjson
from django.conf import settings
from django.template.backends.django import Template
from django.template.response import TemplateResponse
from lxml import html

from django_unicorn.decorators import timed
from django_unicorn.errors import (
    MissingComponentElementError,
    MissingComponentViewElementError,
    MultipleRootComponentElementError,
    NoRootComponentElementError,
)
from django_unicorn.settings import get_minify_html_enabled, get_script_location
from django_unicorn.utils import generate_checksum, html_element_to_string, sanitize_html

logger = logging.getLogger(__name__)


# https://developer.mozilla.org/en-US/docs/Glossary/Empty_element
EMPTY_ELEMENTS = (
    "<area>",
    "<base>",
    "<br>",
    "<col>",
    "<embed>",
    "<hr>",
    "<img>",
    "<input>",
    "<link>",
    "<meta>",
    "<param>",
    "<source>",
    "<track>",
    "<wbr>",
)


def is_html_well_formed(html_content: str) -> bool:
    """
    Whether the passed-in HTML is missing any closing elements which can cause issues when rendering.
    """

    tag_list = re.split("(<[^>!]*>)", html_content)[1::2]
    stack: deque[str] = deque()

    for tag in tag_list:
        if "/" not in tag:
            cleaned_tag = re.sub(r"(<([\w\-]+)[^>!]*>)", r"<\2>", tag)

            if cleaned_tag not in EMPTY_ELEMENTS:
                stack.append(cleaned_tag)
        elif len(stack) > 0 and (tag.replace("/", "") == stack[len(stack) - 1]):
            stack.pop()

    return len(stack) == 0


def get_root_element(content: str) -> html.HtmlElement:
    """Gets the first tag element for the component or the first element with a `unicorn:view` attribute for a direct
    view.

    Returns:
        lxml element.

        Raises `Exception` if an element cannot be found.
    """
    try:
        if isinstance(content, html.HtmlElement):
            return content

        if "<html" in content.lower():
            root_element = html.fromstring(content)
        else:
            # lxml.html.fragments_fromstring returns a list of elements/strings
            fragments = html.fragments_fromstring(content)
            elements = [f for f in fragments if isinstance(f, html.HtmlElement)]

            if not elements:
                raise MissingComponentElementError("No root element for the component was found")

            root_element = elements[0]
    except MissingComponentViewElementError:
        # Re-raise this specific error
        raise
    except Exception as e:
        if isinstance(e, MissingComponentElementError):
            raise
        raise MissingComponentElementError(f"Failed to parse component HTML: {e}") from e

    if root_element.tag == "html":
        # Check for unicorn:view in descendants
        view_element = None
        for element in root_element.iter():
            if "unicorn:view" in element.attrib or "u:view" in element.attrib:
                view_element = element
                break

        if view_element is None:
            raise MissingComponentViewElementError(
                "An element with an `unicorn:view` attribute is required for a direct view"
            )
        return view_element

    return root_element


def assert_has_single_wrapper_element(content: str, component_name: str) -> None:
    """Assert that there is only one root element."""
    try:
        if isinstance(content, html.HtmlElement):
            elements = [content]
        else:
            fragments = html.fragments_fromstring(content)
            elements = [f for f in fragments if isinstance(f, html.HtmlElement)]
    except Exception:
        # Should have been caught by get_root_element usually
        return

    # Check if any element has unicorn:view or u:view - these are direct views
    # and should skip the multiple root check
    for element in elements:
        has_view = "unicorn:view" in element.attrib or "u:view" in element.attrib
        has_poll = "unicorn:poll" in element.attrib or "u:poll" in element.attrib

        if has_view or (has_view and has_poll):
            # If the root element is a direct view (with or without poll), skip the check
            return

    if len(elements) > 1:
        raise MultipleRootComponentElementError(
            f"The '{component_name}' component appears to have multiple root elements."
        )

    if not elements:
        raise NoRootComponentElementError(f"The '{component_name}' component does not appear to have one root element.")

    if f"<{elements[0].tag}>" in EMPTY_ELEMENTS:
        raise NoRootComponentElementError(
            f"The '{component_name}' component root element cannot be a void element like <{elements[0].tag}>. "
            "It must be a wrapping element."
        )


class UnicornTemplateResponse(TemplateResponse):
    def __init__(
        self,
        template,
        request,
        *,
        context=None,
        content_type=None,
        status=None,
        charset=None,
        using=None,
        component=None,
        init_js=False,
        **kwargs,  # noqa: ARG002
    ):
        super().__init__(
            template=template,
            request=request,
            context=context,
            content_type=content_type,
            status=status,
            charset=charset,
            using=using,
        )

        self.component = component
        self.init_js = init_js

    def resolve_template(self, template):
        """Override the TemplateResponseMixin to resolve a list of Templates.

        Calls the super which accepts a template object, path-to-template, or list of paths if the first
        object in the sequence is not a Template.
        """

        if isinstance(template, list | tuple):
            if isinstance(template[0], Template):
                return template[0]

        return super().resolve_template(template)

    @timed
    def render(self):
        response = super().render()

        if not self.component or not self.component.component_id:
            return response

        content = response.content.decode("utf-8")

        # Only check if HTML is well-formed in DEBUG mode
        if settings.DEBUG and not is_html_well_formed(content):
            logger.warning(
                f"The HTML in '{self.component.component_name}' appears to be missing a closing tag. "
                "That can potentially cause errors in Unicorn."
            )

        try:
            assert_has_single_wrapper_element(content, self.component.component_name)
        except (NoRootComponentElementError, MultipleRootComponentElementError) as ex:
            logger.warning(ex)

        root_element = get_root_element(content)

        # Prepare Data
        frontend_context_variables = self.component.get_frontend_context_variables()
        frontend_context_variables_dict = orjson.loads(frontend_context_variables)
        checksum = generate_checksum(frontend_context_variables_dict)

        # Modify Attributes
        root_element.set("unicorn:id", self.component.component_id)
        if hasattr(self.component, "component_name"):
            root_element.set("unicorn:name", self.component.component_name)
        root_element.set("unicorn:key", str(self.component.component_key or ""))
        root_element.set("unicorn:checksum", checksum)
        root_element.set("unicorn:data", frontend_context_variables)
        root_element.set("unicorn:calls", orjson.dumps(self.component.calls).decode("utf-8"))

        # Calculate content hash (without script)
        rendered_template_no_script = html_element_to_string(root_element)
        content_hash = generate_checksum(rendered_template_no_script)

        rendered_template = rendered_template_no_script

        # Inject Scripts
        if self.init_js:
            init = {
                "id": self.component.component_id,
                "name": self.component.component_name,
                "key": self.component.component_key,
                "data": orjson.loads(frontend_context_variables),
                "calls": self.component.calls,
                "hash": content_hash,
            }
            init_json = orjson.dumps(init).decode("utf-8")
            init_json_safe = sanitize_html(init_json)
            json_element_id = f"unicorn:data:{self.component.component_id}"
            init_script = (
                f"Unicorn.componentInit(JSON.parse(document.getElementById('{json_element_id}').textContent));"
            )

            # Create JSON script tag
            json_tag = html.Element("script", type="application/json", id=json_element_id)
            json_tag.text = init_json_safe

            if self.component.parent:
                self.component._init_script = init_script
                self.component._json_tag = json_tag
            else:
                json_tags = [json_tag]

                descendants = []
                descendants.append(self.component)
                while descendants:
                    descendant = descendants.pop()
                    for child in descendant.children:
                        init_script = f"{init_script} {child._init_script}"

                        if hasattr(child, "_json_tag"):
                            json_tags.append(child._json_tag)
                            del child._json_tag

                        descendants.append(child)

                script_tag = html.Element("script", type="module")
                script_content = (
                    "if (typeof Unicorn === 'undefined') { "
                    "console.error('Unicorn is missing. Do you need {% load unicorn %} or {% unicorn_scripts %}?') "
                    "} else { " + init_script + " }"
                )
                script_tag.text = script_content

                if get_script_location() == "append":
                    root_element.append(script_tag)
                    for t in json_tags:
                        root_element.append(t)
                    rendered_template = html_element_to_string(root_element)
                else:
                    root_html = html_element_to_string(root_element)
                    script_html = html_element_to_string(script_tag)
                    for t in json_tags:
                        script_html += html_element_to_string(t)
                    rendered_template = root_html + script_html

        self.component.rendered(rendered_template)
        response.content = rendered_template

        if get_minify_html_enabled():
            from htmlmin import minify  # noqa: PLC0415  # type: ignore

            minified_html = minify(response.content.decode())
            if len(minified_html) < len(rendered_template):
                response.content = minified_html

        return response



import logging
from functools import wraps

from django.http import HttpRequest, JsonResponse
from django.http.response import HttpResponseNotModified
from django.views.decorators.csrf import csrf_protect, ensure_csrf_cookie
from django.views.decorators.http import require_POST

from django_unicorn.decorators import timed
from django_unicorn.errors import RenderNotModifiedError, UnicornViewError
from django_unicorn.views.message import UnicornMessageHandler
from django_unicorn.views.request import ComponentRequest

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def handle_error(view_func):
    """
    Returns a JSON response with an error if necessary.
    """

    def wrapped_view(*args, **kwargs):
        try:
            return view_func(*args, **kwargs)
        except UnicornViewError as e:
            return JsonResponse({"error": str(e)})
        except RenderNotModifiedError:
            return HttpResponseNotModified()
        except AssertionError as e:
            return JsonResponse({"error": str(e)})

    return wraps(view_func)(wrapped_view)


@timed
@handle_error
@ensure_csrf_cookie
@csrf_protect  # type: ignore
@require_POST  # type: ignore
def message(request: HttpRequest, component_name: str | None = None) -> JsonResponse:  # type: ignore
    """
    Endpoint that instantiates the component and does the correct action
    (set an attribute or call a method) depending on the JSON payload in the body.

    Args:
        param request: HttpRequest for the function-based view.
        param: component_name: Name of the component, e.g. "hello-world".

    Returns:
        `JsonRequest` with the following structure in the body:
        {
        "id": component_id,
        "dom": html,  # re-rendered version of the component after actions in the payload are completed
        "data": {},  # updated data after actions in the payload are completed
        "errors": {},  # form validation errors
        "return": {}, # optional return value from an executed action
        "parent": {}  # optional representation of the parent component
        }
    """

    if not component_name:
        raise AssertionError("Missing component name in url")

    component_request = ComponentRequest(request, component_name)
    handler = UnicornMessageHandler(request)
    json_result = handler.handle(component_request)

    return JsonResponse(json_result, json_dumps_params={"separators": (",", ":")})



from django_unicorn.call_method_parser import parse_call_method_name


def setup_function():
    """Clears lru_cache before every test in the module."""
    parse_call_method_name.cache_clear()


def test_single_int_arg():
    expected = ("set_name", (1,), {})
    actual = parse_call_method_name("set_name(1)")

    assert actual == expected


def test_single_dict_arg():
    expected = ("set_name", ({"1": 2, "3": 4},), {})
    actual = parse_call_method_name('set_name({"1": 2, "3": 4})')

    assert actual == expected


def test_multiple_args():
    expected = ("set_name", (0, {"1": 2}), {})
    actual = parse_call_method_name('set_name(0, {"1": 2})')

    assert actual == expected


def test_multiple_args_2():
    expected = ("set_name", (1, 2), {})
    actual = parse_call_method_name("set_name(1, 2)")

    assert actual == expected


def test_var_with_curly_braces():
    expected = (
        "set_name",
        ("{}",),
        {},
    )
    actual = parse_call_method_name('set_name("{}")')

    assert actual == expected


def test_one_arg():
    expected = (
        "set_name",
        ("1",),
        {},
    )
    actual = parse_call_method_name('set_name("1")')

    assert actual == expected


def test_kwargs():
    expected = ("set_name", (), {"kwarg1": "wow"})
    actual = parse_call_method_name("set_name(kwarg1='wow')")

    assert actual == expected


def test_args_and_kwargs():
    expected = ("set_name", (8, 9), {"kwarg1": "wow"})
    actual = parse_call_method_name("set_name(8, 9, kwarg1='wow')")

    assert actual == expected


def test_special_method_without_parens():
    expected = ("$reset", (), {})
    actual = parse_call_method_name("$reset")

    assert actual == expected


def test_special_method_with_parens():
    expected = ("$reset", (), {})
    actual = parse_call_method_name("$reset()")

    assert actual == expected


def test_string_with_single_quotes():
    expected = ("set_name", ("It's a test",), {})
    actual = parse_call_method_name("set_name('It\\'s a test')")

    assert actual == expected


def test_string_with_double_quotes():
    expected = ("set_name", ('He said "hello"',), {})
    actual = parse_call_method_name("set_name('He said \"hello\"')")

    assert actual == expected


def test_string_with_backslashes():
    expected = ("set_name", ("C:\\Users\\test",), {})
    actual = parse_call_method_name("set_name('C:\\\\Users\\\\test')")

    assert actual == expected


def test_complex_chemical_name_issue_607():
    """Test for issue #607: string values with single quotes break function calls"""
    chemical_name = "Chloro(2-dicyclohexylphosphino-3,6-dimethoxy-2',4',6'-tri-i-propyl-1,1'-biphenyl)(2'-amino-1,1'-biphenyl-2-yl)palladium(II)"  # noqa: E501
    expected = ("set_new_catalyst", (chemical_name,), {})
    actual = parse_call_method_name(
        "set_new_catalyst('Chloro(2-dicyclohexylphosphino-3,6-dimethoxy-2\\',"
        "4\\',6\\'-tri-i-propyl-1,1\\'-biphenyl)(2\\'-amino-1,1\\'-biphenyl-2-yl)palladium(II)')"
    )

    assert actual == expected


def test_string_with_mixed_quotes_and_parentheses():
    expected = ("set_name", ("test(with 'quotes' and \"double\")",), {})
    actual = parse_call_method_name("set_name('test(with \\'quotes\\' and \"double\")')")

    assert actual == expected



import pytest
from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView
from example.coffee.models import Flavor, Taste


class M2MComponent(UnicornView):
    template_name = "templates/test_component.html"
    taste: Taste | None = None

    def mount(self):
        if self.component_args:
            self.taste = Taste.objects.get(pk=self.component_args[0])

    def refresh(self):
        # usage: unicorn:poll="refresh"
        if self.taste:
            self.taste = Taste.objects.get(pk=self.taste.pk)


@pytest.mark.django_db
def test_m2m_overwriting(client):
    # Setup
    flavor1 = Flavor.objects.create(name="Flavor 1")
    flavor2 = Flavor.objects.create(name="Flavor 2")
    taste = Taste.objects.create(name="Test Taste")
    taste.flavor.add(flavor1)

    # Initial state (client side)
    # We strictly define what we expect the client state to resemble after initial render
    # 'flavor' is M2M. Unicorn serialization of M2M depends on configuration/implementation.
    # Attempting to mimic the "stale" state on the client.
    # If Unicorn serializes M2M as list of PKs:
    data = {"taste": {"pk": taste.pk, "flavor": [flavor1.pk]}}

    # 2. Modify DB (Admin action) - Add Flavor 2
    taste.flavor.add(flavor2)
    assert taste.flavor.count() == 2

    # 3. Component Update (Poll/Refresh)
    # Client sends back the OLD data (only flavor1 in the payload)
    # This simulates the frontend sending back its known state which is outdated.
    action_queue = [{"payload": {"name": "refresh"}, "type": "callMethod"}]

    response = post_and_get_response(
        client,
        url="/message/tests.views.test_m2m_overwriting.M2MComponent",
        data=data,
        action_queue=action_queue,
    )

    assert not response.get("errors")

    # 4. Assert DB state
    taste.refresh_from_db()

    # If the bug exists, Flavor 2 might vary well be gone because the component
    # restored the 'taste' model from the 'data' (which only had flavor1)
    # and saved it or synced the M2M.
    assert taste.flavor.count() == 2, "DB M2M changes were overwritten!"
    assert flavor2 in taste.flavor.all()



from dataclasses import dataclass
from decimal import Decimal

from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView


@dataclass
class ExampleDataclass:
    hello: str = "world"


class FakeObjectsComponent(UnicornView):
    template_name = "templates/test_component.html"

    decimal_example: Decimal = Decimal("1.1")
    float_example: float = 1.2
    int_example: int = 3
    dataclass_example: ExampleDataclass | None = None

    def assert_int(self):
        assert self.int_example == 4

    def assert_float(self):
        assert self.float_example == 1.3

    def assert_decimal(self):
        assert self.decimal_example == Decimal("1.5")

    def assert_dataclass(self):
        assert self.dataclass_example is not None
        assert self.dataclass_example == ExampleDataclass(hello="world")
        assert self.dataclass_example.hello == "world"


FAKE_OBJECTS_COMPONENT_URL = "/message/tests.views.message.test_type_hints.FakeObjectsComponent"


def test_message_int(client):
    data = {"int_example": "4"}
    action_queue = [
        {
            "payload": {"name": "assert_int"},
            "type": "callMethod",
        }
    ]
    response = post_and_get_response(
        client,
        url=FAKE_OBJECTS_COMPONENT_URL,
        data=data,
        action_queue=action_queue,
    )

    assert not response.get("error")  # UnicornViewError/AssertionError returns a a JsonResponse with "error" key
    assert not response["errors"]


def test_message_float(client):
    data = {"float_example": "1.3"}
    action_queue = [
        {
            "payload": {"name": "assert_float"},
            "type": "callMethod",
        }
    ]
    response = post_and_get_response(
        client,
        url=FAKE_OBJECTS_COMPONENT_URL,
        data=data,
        action_queue=action_queue,
    )

    assert not response.get("error")  # UnicornViewError/AssertionError returns a a JsonResponse with "error" key
    assert not response["errors"]


def test_message_decimal(client):
    data = {"decimal_example": "1.5"}
    action_queue = [
        {
            "payload": {"name": "assert_decimal"},
            "type": "callMethod",
        }
    ]
    response = post_and_get_response(
        client,
        url=FAKE_OBJECTS_COMPONENT_URL,
        data=data,
        action_queue=action_queue,
    )

    assert not response.get("error")  # UnicornViewError/AssertionError returns a a JsonResponse with "error" key
    assert not response["errors"]


def test_message_dataclass(client):
    data = {"dataclass_example": {"hello": "world"}}
    action_queue = [
        {
            "payload": {"name": "assert_dataclass"},
            "type": "callMethod",
        }
    ]
    response = post_and_get_response(
        client,
        url=FAKE_OBJECTS_COMPONENT_URL,
        data=data,
        action_queue=action_queue,
    )

    assert not response.get("error")  # UnicornViewError/AssertionError returns a a JsonResponse with "error" key
    assert not response["errors"]



import logging
from warnings import warn

from django.conf import settings

logger = logging.getLogger(__name__)


SETTINGS_KEY = "UNICORN"
LEGACY_SETTINGS_KEY = f"DJANGO_{SETTINGS_KEY}"

DEFAULT_MORPHER_NAME = "morphdom"
MORPHER_NAMES = (
    "morphdom",
    "alpine",
)


def get_settings():
    unicorn_settings = {}

    if hasattr(settings, LEGACY_SETTINGS_KEY):
        # TODO: Log deprecation message here
        unicorn_settings = getattr(settings, LEGACY_SETTINGS_KEY)
    elif hasattr(settings, SETTINGS_KEY):
        unicorn_settings = getattr(settings, SETTINGS_KEY)

    return unicorn_settings


def get_setting(key, default=None):
    unicorn_settings = get_settings()

    return unicorn_settings.get(key, default)


def get_serial_settings():
    return get_setting("SERIAL", {})


def get_cache_alias():
    return get_setting("CACHE_ALIAS", "default")


def get_morpher_settings():
    options = get_setting("MORPHER", {"NAME": DEFAULT_MORPHER_NAME})

    # Legacy `RELOAD_SCRIPT_ELEMENTS` setting that needs to go to `MORPHER.RELOAD_SCRIPT_ELEMENTS`
    reload_script_elements = get_setting("RELOAD_SCRIPT_ELEMENTS")

    if reload_script_elements:
        msg = 'The `RELOAD_SCRIPT_ELEMENTS` setting is deprecated. Use \
`MORPHER["RELOAD_SCRIPT_ELEMENTS"]` instead.'
        warn(msg, DeprecationWarning, stacklevel=1)

        options["RELOAD_SCRIPT_ELEMENTS"] = reload_script_elements

    if not options.get("NAME"):
        options["NAME"] = DEFAULT_MORPHER_NAME

    morpher_name = options["NAME"]

    if not morpher_name or morpher_name not in MORPHER_NAMES:
        raise AssertionError(f"Unknown morpher name: {morpher_name}")

    return options


def get_script_location():
    """
    Valid choices: "append", "after". Default is "after".
    """

    return get_setting("SCRIPT_LOCATION", "after")


def get_serial_enabled():
    """
    Default serial enabled is `False`.
    """
    enabled = get_serial_settings().get("ENABLED", False)

    if enabled and settings.CACHES:
        cache_alias = get_cache_alias()
        cache_settings = settings.CACHES.get(cache_alias, {})
        cache_backend = cache_settings.get("BACKEND")

        if cache_backend == "django.core.cache.backends.dummy.DummyCache":
            return False

    return enabled


def get_serial_timeout():
    """
    Default serial timeout is 60 seconds.
    """
    return get_serial_settings().get("TIMEOUT", 60)


def get_minify_html_enabled():
    minify_html_enabled = get_setting("MINIFY_HTML", False)

    if minify_html_enabled:
        try:
            import htmlmin  # type: ignore # noqa: F401, PLC0415
        except ModuleNotFoundError:
            logger.error(
                "MINIFY_HTML is `True`, but minify extra could not be imported. Install with `django-unicorn[minify]`."
            )

            return False

    return minify_html_enabled



from django.utils.functional import cached_property

from django_unicorn.components.unicorn_view import UnicornView, constructed_views_cache


class PropertyView(UnicornView):
    class Meta:
        javascript_exclude = ("user_id", "cached_user_id")

    @property
    def user_id(self):
        return self.request.user.id

    @cached_property
    def cached_user_id(self):
        return self.request.user.id


class User:
    id = 1


def test_issue_650_property_access_on_restore(settings, rf):
    request = rf.get("/")
    request.user = User()

    # Register component
    settings.UNICORN = {**settings.UNICORN, "COMPONENTS": {"property-view": PropertyView}}

    # 1. Create component
    component_id = "test-issue-650"
    component = UnicornView.create(component_id=component_id, component_name="property-view", request=request)

    assert component.user_id == 1
    assert component.cached_user_id == 1

    # Clear memory cache to force restore from backend cache (pickle)
    constructed_views_cache.clear()

    # 2. Restore component
    # We use the same request logic as if it was a new request
    request2 = rf.get("/")
    request2.user = User()

    component_restored = UnicornView.create(component_id=component_id, component_name="property-view", request=request2)

    # This fails if request is not set on restored component
    assert component_restored.user_id == 1
    # This also fails if request is not set AND it wasn't validly cached
    assert component_restored.cached_user_id == 1



import pytest

from django_unicorn.components.unicorn_template_response import (
    assert_has_single_wrapper_element,
    get_root_element,
)
from django_unicorn.errors import MultipleRootComponentElementError


def test_issue_668_unicorn_view_and_poll():
    # User reported case
    html = '<div unicorn:view unicorn:poll="get_updates">Content</div>'
    root_element = get_root_element(html)

    # This should NOT raise MultipleRootComponentElementError
    try:
        assert_has_single_wrapper_element(root_element, "test-component")
    except MultipleRootComponentElementError:
        pytest.fail("MultipleRootComponentElementError raised incorrectly for unicorn:view + unicorn:poll")


def test_issue_668_unicorn_view_no_poll():
    # Control case
    html = "<div unicorn:view>Content</div>"
    root_element = get_root_element(html)

    assert_has_single_wrapper_element(root_element, "test-component")



from datetime import datetime

import pytest

from django_unicorn.components import UnicornView
from django_unicorn.views.action_parsers.utils import set_property_value
from example.coffee.models import Flavor, Taste


class FakeComponent(UnicornView):
    string = "property_view"
    integer = 99
    datetime = datetime(2020, 1, 1)  # noqa: DTZ001
    array: list[str] = []  # noqa: RUF012
    model = Flavor(name="initial-flavor")
    queryset = Flavor.objects.none()
    taste = Taste(name="bitter")


def test_set_property_value_str():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_str")
    assert "property_view" == component.string

    string = "property_view_updated"
    data = {"string": "property_view_updated"}

    set_property_value(
        component,
        "string",
        string,
        data,
    )

    assert component.string == string
    assert data["string"] == string


def test_set_property_value_int():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_int")
    assert 99 == component.integer

    integer = 100
    data = {"integer": None}

    set_property_value(component, "integer", integer, data)

    assert component.integer == integer
    assert data["integer"] == integer


def test_set_property_value_datetime():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_datetime")
    assert datetime(2020, 1, 1) == component.datetime  # noqa: DTZ001

    dt = datetime(2020, 1, 2)  # noqa: DTZ001
    data = {"datetime": None}

    set_property_value(component, "datetime", dt, data)

    assert component.datetime == dt
    assert data["datetime"] == dt


def test_set_property_value_model():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_model")
    assert "initial-flavor" == component.model.name

    model = Flavor(name="test-flavor")
    data = {"model": {}}

    set_property_value(
        component,
        "model",
        model,
        data,
    )

    assert component.model.name == model.name
    assert data["model"] == model


@pytest.mark.django_db
def test_set_property_value_queryset():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_queryset")
    assert len(component.queryset) == 0

    flavor_one = Flavor(name="test-flavor-one")
    flavor_one.save()
    flavor_two = Flavor(name="test-flavor-two")
    flavor_two.save()
    queryset = Flavor.objects.all()[:2]
    data = {"queryset": []}

    set_property_value(
        component,
        "queryset",
        queryset,
        data,
    )

    assert len(queryset) == 2
    assert data["queryset"] == queryset


@pytest.mark.django_db
def test_set_property_value_many_to_many_is_referenced():
    component = FakeComponent(component_name="test", component_id="test_set_property_value_many_to_many_is_referenced")
    component.model.save()
    assert component.model.taste_set.count() == 0  # type: ignore

    taste = Taste(name="Bitter")
    taste.save()
    flavor = Flavor(name="test-flavor")
    flavor.save()
    flavor.taste_set.add(taste)  # type: ignore

    data = {"model": {}}

    set_property_value(
        component,
        "model.taste_set",
        [taste.pk],
        data,
    )

    assert data["model"]["taste_set"] == [taste.pk]
    assert component.model.taste_set.count() == 1  # type: ignore


@pytest.mark.django_db
def test_set_property_value_many_to_many_references_model():
    component = FakeComponent(
        component_name="test", component_id="test_set_property_value_many_to_many_references_model"
    )
    component.taste.save()
    assert component.taste.flavor.count() == 0  # type: ignore

    taste = Taste(name="Bitter")
    taste.save()
    flavor = Flavor(name="test-flavor")
    flavor.save()
    flavor.taste_set.add(taste)  # type: ignore

    data = {"taste": {}}

    set_property_value(
        component,
        "taste.flavor",
        [flavor.pk],
        data,
    )

    assert data["taste"]["flavor"] == [flavor.pk]
    assert component.taste.flavor.count() == 1  # type: ignore



from tests.views.fake_components import FakeComponent

from django_unicorn.views.action_parsers.call_method import _get_property_value


def test_get_property_value():
    component = FakeComponent(component_name="test", component_id="asdf")

    component.check = False
    check_value = _get_property_value(component, "check")
    assert check_value is False

    component.check = True
    check_value = _get_property_value(component, "check")

    assert check_value is True


def test_get_property_value_nested():
    component = FakeComponent(component_name="test", component_id="asdf")

    component.nested["check"] = False
    check_value = _get_property_value(component, "nested.check")
    assert check_value is False

    component.nested["check"] = True
    check_value = _get_property_value(component, "nested.check")

    assert check_value is True



import pytest

from django_unicorn.components.unicorn_template_response import (
    assert_has_single_wrapper_element,
    get_root_element,
)
from django_unicorn.errors import (
    MissingComponentElementError,
    MissingComponentViewElementError,
    MultipleRootComponentElementError,
    NoRootComponentElementError,
)
from django_unicorn.utils import html_element_to_string


def test_get_root_element():
    expected = "<div>test</div>"
    component_html = "<div>test</div>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_with_comment():
    expected = "<div>test</div>"
    component_html = "<!-- some comment --><div>test</div>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_with_blank_string():
    expected = "<div>test</div>"
    component_html = "\n<div>test</div>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_no_element():
    component_html = "\n"
    with pytest.raises(MissingComponentElementError):
        get_root_element(component_html)


def test_get_root_element_direct_view_unicorn():
    # lxml serializes empty attributes as just the attribute name
    expected = "<div unicorn:view>test</div>"
    component_html = "<html><head></head><body><div unicorn:view>test</div></body></html>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_direct_view_u():
    # lxml serializes empty attributes as just the attribute name
    expected = "<div u:view>test</div>"
    component_html = "<html><head></head><body><div u:view>test</div></body></html>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_as_direct_view_unicorn():
    expected = "<div unicorn:view>test</div>"
    component_html = "<div unicorn:view>test</div>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_as_direct_view_u():
    expected = "<div u:view>test</div>"
    component_html = "<div u:view>test</div>"
    actual = get_root_element(component_html)
    assert html_element_to_string(actual) == expected


def test_get_root_element_direct_view_no_view():
    component_html = "<html><head></head><body><div>test</div></body></html>"
    with pytest.raises(MissingComponentViewElementError):
        get_root_element(component_html)


def test_assert_has_single_wrapper_element_one_element_no_children():
    html_content = '<input unicorn:model="name">'
    with pytest.raises(NoRootComponentElementError):
        assert_has_single_wrapper_element(html_content, "test-component-name")


def test_assert_has_single_wrapper_element_multiple_roots():
    html_content = """<div>
    <input unicorn:model="name1">
</div>
<input unicorn:model="name2">"""

    with pytest.raises(MultipleRootComponentElementError):
        assert_has_single_wrapper_element(html_content, "test-component-name")


def test_assert_has_single_wrapper_element_true():
    html_content = '<div><input unicorn:model="name"></div>'
    assert_has_single_wrapper_element(html_content, "test-component-name")
